package com.prosc.commerce.gateway.linkpoint;

import com.prosc.commerce.gateway.*;
import com.prosc.commerce.gateway.linkpoint.ws.*;
import org.w3c.dom.Document;
import org.w3c.dom.NodeList;
import org.xml.sax.InputSource;
import org.xml.sax.SAXException;

import javax.net.ssl.HttpsURLConnection;
import javax.net.ssl.KeyManager;
import javax.net.ssl.KeyManagerFactory;
import javax.net.ssl.SSLContext;
import javax.xml.parsers.DocumentBuilder;
import javax.xml.parsers.DocumentBuilderFactory;
import javax.xml.parsers.ParserConfigurationException;
import javax.xml.ws.Binding;
import javax.xml.ws.BindingProvider;
import javax.xml.ws.handler.Handler;
import javax.xml.ws.soap.SOAPFaultException;
import java.io.FileInputStream;
import java.io.IOException;
import java.io.StringReader;
import java.math.BigInteger;
import java.security.GeneralSecurityException;
import java.security.KeyStore;
import java.security.KeyStoreException;
import java.security.NoSuchAlgorithmException;
import java.security.cert.CertificateException;
import java.text.DateFormat;
import java.text.SimpleDateFormat;
import java.util.Date;
import java.util.List;
import java.util.logging.Level;
import java.util.logging.Logger;

/**
 * Created by IntelliJ IDEA.
 * User: sterling
 * Date: 6/3/11
 * Time: 2:52 AM
 * To change this template use File | Settings | File Templates.
 */
public class FirstDataGateway2 extends CreditCardPaymentGateway {
	private static final Logger log = Logger.getLogger(FirstDataGateway2.class.getName());
	private static final DateFormat monthFormat = new SimpleDateFormat("MM");
	private static final DateFormat yearFormat = new SimpleDateFormat("yy");
	private static final DateFormat recurringFormat = new SimpleDateFormat("yyyyMMdd");

	private static FDGGWSApiOrderService service;

	private static final String PRODUCTION_URL = "https://ws.firstdataglobalgateway.com/fdggwsapi/services";
	private static final String MODIFY_FUNCTION = "modify";
	private static final String CANCEL_FUNCTION = "cancel";

	public static final String TEST_URL = "https://ws.merchanttest.firstdataglobalgateway.com/fdggwsapi/services";
	public static enum AccountType {
		PRIMARY_CHECKING("PC"), PRIMARY_SAVINGS("PS"), BACKUP_CHECKING("BC"), BACKUP_SAVINGS("BS");

		String accountTypeAbbreviation;

		AccountType(String accountTypeAbbreviation) {
			this.accountTypeAbbreviation = accountTypeAbbreviation;
		}

		public String getAccountTypeAbbreviation() {
			return accountTypeAbbreviation;
		}

		public static String valuesAsString() {
			StringBuilder builder = new StringBuilder("");
			AccountType[] values = AccountType.values();
			int i = 0;
			for (AccountType value : values) {
				boolean last = values.length - 1 == i;
				if (last) {
					builder.append(value.name());
				} else {
					builder.append(value.name()).append(", ");
				}
				i++;
			}
			return builder.toString();

		}
	}

	private Date recurringStartDate;
	private Integer recurringNumberInstallments;
	private BigInteger recurringInstallmentFrequency;
	private String recurringInstallmentPeriod;
	private Integer recurringMaxFailures;
	private String accountName;
	private String accountNumber;
	private String accountHolderName;
	private String abaCode;
	private AccountType accountType;
	private String bankName;
	private String bankState;
	private String checkNumber;
	private String driversLicenseNumber;
	private String driversLicenseState;
	private String ssn;

	private boolean isEcheck = false;

	public void setRecurringInstallmentFrequency(BigInteger recurringInstallmentFrequency) {
		this.recurringInstallmentFrequency = recurringInstallmentFrequency;
	}

	public void setRecurringInstallmentPeriod(String recurringInstallmentPeriod) {
		this.recurringInstallmentPeriod = recurringInstallmentPeriod;
	}

	public void setRecurringMaxFailures(Integer recurringMaxFailures) {
		this.recurringMaxFailures = recurringMaxFailures;
	}

	public void setRecurringNumberInstallments(Integer recurringNumberInstallments) {
		this.recurringNumberInstallments = recurringNumberInstallments;
	}

	public void setRecurringStartDate(Date recurringStartDate) {
		this.recurringStartDate = recurringStartDate;
	}

	public void setEcheck(boolean isEcheck) {
		this.isEcheck = isEcheck;
	}

	public Date getRecurringStartDate() {
		return recurringStartDate;
	}

	public Integer getRecurringNumberInstallments() {
		return recurringNumberInstallments;
	}

	public BigInteger getRecurringInstallmentFrequency() {
		return recurringInstallmentFrequency;
	}

	public String getRecurringInstallmentPeriod() {
		return recurringInstallmentPeriod;
	}

	public Integer getRecurringMaxFailures() {
		return recurringMaxFailures;
	}

	public boolean isEcheck() {
		return isEcheck;
	}

	public static enum TransactionTypeEnum {
		AuthCaptureTxn("sale"), AuthorizeTxn("preAuth"), CaptureTxn ("postAuth"), ForceCaptureTxn("forceTicket"), RefundTxn("return"), VoidTxn("void");

		String transactionType;

		TransactionTypeEnum(String transactionType) {
			this.transactionType = transactionType;
		}

		public String getTransactionType() {
			return transactionType;
		}
	}

	public static enum CardCodeIndicator {
		PROVIDED, NOT_PROVIDED, ILLEGIBLE, NO_IMPRINT, NOT_PRESENT;

		public static String valuesAsString() {
			StringBuilder builder = new StringBuilder("");
			CardCodeIndicator[] values = CardCodeIndicator.values();
			int i = 0;
			for (CardCodeIndicator value : values) {
				boolean last = values.length - 1 == i;
				if (last) {
					builder.append(value.name());
				} else {
					builder.append(value.name()).append(", ");
				}
				i++;
			}
			return builder.toString();

		}
	}

	private TransactionTypeEnum transactionType = TransactionTypeEnum.AuthCaptureTxn;
	private CardCodeIndicator cardCodeIndicator;
	private String tDate;
	private String lastFaultMessage;

	FDGGWSApiOrderResponse orderResponse;
	FDGGWSApiActionResponse actionResponse;

	private String certificatePath;
	private String certificatePassword;

	/**
	 * Sets the client certificate to be used in all subsequent Web service operations.
	 * You must call this function before performing any Web service operations. Once any Web service operation has been
	 * executed, this function will have no effect unless FileMaker is restarted.
	 * @param certificatePath
	 * @param certificatePassword
	 */
	public void setCertificate(String certificatePath, String certificatePassword) {
		this.certificatePath = certificatePath;
		this.certificatePassword = certificatePassword;
	}

	private void _setCertificate()
			throws KeyStoreException, IOException, NoSuchAlgorithmException, CertificateException {
		log.fine("Certificate path: " + certificatePath);
		log.fine("Certificate password: " + certificatePassword);
		System.setProperty("sun.security.ssl.allowUnsafeRenegotiation", "true");
		if (certificatePath != null) {
			CertUtils.setKeystore(certificatePath, certificatePassword == null ? "" : certificatePassword);
		}
	}

	public TransactionTypeEnum getTransactionType() {
		return transactionType;
	}

	public void setTransactionType(TransactionTypeEnum transactionType) {
		this.transactionType = transactionType;
	}

	public CardCodeIndicator getCardCodeIndicator() {
		return cardCodeIndicator;
	}

	public void setCardCodeIndicator(CardCodeIndicator cardCodeIndicator) {
		this.cardCodeIndicator = cardCodeIndicator;
	}

	public String gettDate() {
		return tDate;
	}

	public void settDate(String tDate) {
		this.tDate = tDate;
	}

	public static FDGGWSApiOrderService getService() {
		if(service == null) {
			service = new FDGGWSApiOrderService();
		}
		return service;
	}

	private void initializeSecuritySettings() throws GeneralSecurityException, IOException {
		KeyManagerFactory keyManagerFactory = KeyManagerFactory.getInstance(KeyManagerFactory.getDefaultAlgorithm());
		KeyStore keyStore = KeyStore.getInstance(KeyStore.getDefaultType());

		if(certificatePath != null && certificatePassword != null) {
			log.fine("Configuring trust manager using client-side certificate");
			FileInputStream certStream = new FileInputStream(certificatePath);
			try {
				keyStore.load(certStream, certificatePassword.toCharArray());
				keyManagerFactory.init(keyStore, certificatePassword .toCharArray());
			} finally {
				certStream.close();
			}

			SSLContext sslContext = SSLContext.getInstance("SSL");
			KeyManager[] keyManagers = null;
			try {
				keyManagers = keyManagerFactory.getKeyManagers();
			} catch (IllegalStateException e) {
				String errMsg = "An error occurred while getting key managers from a factory";
				log.log(Level.WARNING, errMsg, e);
			}

			sslContext.init(keyManagers, null, null);
			HttpsURLConnection.setDefaultSSLSocketFactory(sslContext.getSocketFactory());
		} else {
			log.fine("No client-side certificate info provided, so trust manager will not be configured");
		}
	}

	@Override
	public String getLastValue(String key) {
		return null;  //To change body of implemented methods use File | Settings | File Templates.
	}

	private class FirstDataPaymentInterface implements SinglePaymentInterface, RecurringPaymentSupport, ECheckSupport {

		public void authCaptureTxn() throws PaymentAuthorizationException, PaymentIOException {
			try {
				initializeSecuritySettings();
			} catch (Exception e) {
				String errMsg = "An error occurred while initializing the client-side certificate settings";
				log.log(Level.SEVERE, errMsg, e);
			}
			FDGGWSApiOrder port = getService().getFDGGWSApiOrderPort();
			setHandlers((BindingProvider) port);
			FDGGWSApiOrderRequest orderRequest = new FDGGWSApiOrderRequest();

			Transaction transaction = new Transaction();

			createAuthTransaction(transaction, TransactionTypeEnum.AuthCaptureTxn);

			orderRequest.setTransaction(transaction);

			try {
				orderResponse = port.fdggwsApiOrder(orderRequest);
			} catch (SOAPFaultException e) {
				String errMsg = "A soap fault occurred while trying to perform a sale transaction";
				log.log(Level.SEVERE, errMsg, e);
				throw new PaymentAuthorizationException(lastFaultMessage != null ? lastFaultMessage : errMsg);
			}
			if(!orderResponse.getTransactionResult().equals("APPROVED")) {
				throw new PaymentAuthorizationException(orderResponse.getErrorMessage(), getLastResponse());
			}

		}

		public void authorizeTxn() throws PaymentAuthorizationException, PaymentIOException {
			try {
				initializeSecuritySettings();
			} catch (Exception e) {
				String errMsg = "An error occurred while initializing the client-side certificate settings";
				log.log(Level.SEVERE, errMsg, e);
			}
			FDGGWSApiOrder port = getService().getFDGGWSApiOrderPort();
			setHandlers((BindingProvider) port);
			FDGGWSApiOrderRequest orderRequest = new FDGGWSApiOrderRequest();

			Transaction transaction = new Transaction();

			createAuthTransaction(transaction, TransactionTypeEnum.AuthorizeTxn);

			orderRequest.setTransaction(transaction);

			try {
				orderResponse = port.fdggwsApiOrder(orderRequest);
			} catch (SOAPFaultException e) {
				String errMsg = "A soap fault occurred while trying to perform a pre-auth transaction";
				log.log(Level.SEVERE, errMsg, e);
				throw new PaymentAuthorizationException(lastFaultMessage != null ? lastFaultMessage : errMsg);
			}
			if(!orderResponse.getTransactionResult().equals("APPROVED")) {
				throw new PaymentAuthorizationException(orderResponse.getErrorMessage(), getLastResponse());
			}
		}

		public void captureTxn() throws PaymentAuthorizationException, PaymentIOException {
			try {
				initializeSecuritySettings();
			} catch (Exception e) {
				String errMsg = "An error occurred while initializing the client-side certificate settings";
				log.log(Level.SEVERE, errMsg, e);
			}				FDGGWSApiOrder port = getService().getFDGGWSApiOrderPort();
			setHandlers((BindingProvider) port);
			FDGGWSApiOrderRequest orderRequest = new FDGGWSApiOrderRequest();

			Transaction transaction = new Transaction();
			CreditCardTxType txType = new CreditCardTxType();
			txType.setType(TransactionTypeEnum.CaptureTxn.getTransactionType());
			transaction.setCreditCardTxType(txType);

			Payment payment = new Payment();
			payment.setChargeTotal(getAmount());
			transaction.setPayment(payment);

			TransactionDetails transactionDetails = new TransactionDetails();
			transactionDetails.setOrderId(getPreviousTransactionId());
			transaction.setTransactionDetails(transactionDetails);

			Notes notes = new Notes();
			notes.setComments(getChargeDescription());
			transaction.setNotes(notes);

			orderRequest.setTransaction(transaction);
			try {
				orderResponse = port.fdggwsApiOrder(orderRequest);
			} catch (SOAPFaultException e) {
				String errMsg = "A soap fault occurred while trying to perform a post-auth transaction";
				log.log(Level.SEVERE, errMsg, e);
				throw new PaymentAuthorizationException(lastFaultMessage != null ? lastFaultMessage : errMsg);
			}
			if(!orderResponse.getTransactionResult().equals("APPROVED")) {
				throw new PaymentAuthorizationException(orderResponse.getErrorMessage(), getLastResponse());
			}
		}

		public void voidTxn() throws PaymentAuthorizationException, PaymentIOException {
			try {
				initializeSecuritySettings();
			} catch (Exception e) {
				String errMsg = "An error occurred while initializing the client-side certificate settings";
				log.log(Level.SEVERE, errMsg, e);
			}				FDGGWSApiOrder port = getService().getFDGGWSApiOrderPort();
			setHandlers((BindingProvider) port);
			FDGGWSApiOrderRequest orderRequest = new FDGGWSApiOrderRequest();

			Transaction transaction = new Transaction();
			CreditCardTxType txType = new CreditCardTxType();
			txType.setType(TransactionTypeEnum.VoidTxn.getTransactionType());
			transaction.setCreditCardTxType(txType);

			TransactionDetails transactionDetails = new TransactionDetails();
			transactionDetails.setOrderId(getPreviousTransactionId());
			transactionDetails.setTDate(gettDate());
			transaction.setTransactionDetails(transactionDetails);

			Notes notes = new Notes();
			notes.setComments(getChargeDescription());
			transaction.setNotes(notes);

			orderRequest.setTransaction(transaction);
			try {
				orderResponse = port.fdggwsApiOrder(orderRequest);
			} catch (SOAPFaultException e) {
				String errMsg = "A soap fault occurred while trying to perform a void transaction";
				log.log(Level.SEVERE, errMsg, e);
				throw new PaymentAuthorizationException(lastFaultMessage != null ? lastFaultMessage : errMsg);
			}
			if(!orderResponse.getTransactionResult().equals("APPROVED")) {
				throw new PaymentAuthorizationException(orderResponse.getErrorMessage(), getLastResponse());
			}
		}

		public void refundTxn() throws PaymentAuthorizationException, PaymentIOException {
			try {
				initializeSecuritySettings();
			} catch (Exception e) {
				String errMsg = "An error occurred while initializing the client-side certificate settings";
				log.log(Level.SEVERE, errMsg, e);
			}
			FDGGWSApiOrder port = getService().getFDGGWSApiOrderPort();
			setHandlers((BindingProvider) port);
			FDGGWSApiOrderRequest orderRequest = new FDGGWSApiOrderRequest();

			Transaction transaction = new Transaction();
			CreditCardTxType txType = new CreditCardTxType();
			txType.setType(TransactionTypeEnum.RefundTxn.getTransactionType());
			transaction.setCreditCardTxType(txType);

			Payment payment = new Payment();
			payment.setChargeTotal(getAmount());
			transaction.setPayment(payment);

			TransactionDetails transactionDetails = new TransactionDetails();
			transactionDetails.setOrderId(getPreviousTransactionId());
			transaction.setTransactionDetails(transactionDetails);

			Notes notes = new Notes();
			notes.setComments(getChargeDescription());
			transaction.setNotes(notes);

			orderRequest.setTransaction(transaction);
			try {
				orderResponse = port.fdggwsApiOrder(orderRequest);
			} catch (SOAPFaultException e) {
				String errMsg = "A soap fault occurred while trying to perform a refund transaction";
				log.log(Level.SEVERE, errMsg, e);
				throw new PaymentAuthorizationException(lastFaultMessage != null ? lastFaultMessage : errMsg);
			}
			if(!orderResponse.getTransactionResult().equals("APPROVED")) {
				throw new PaymentAuthorizationException(orderResponse.getErrorMessage(), getLastResponse());
			}
		}

		public void cancelSubscription() throws PaymentAuthorizationException {
			try {
				initializeSecuritySettings();
			} catch (Exception e) {
				String errMsg = "An error occurred while initializing the client-side certificate settings";
				log.log(Level.SEVERE, errMsg, e);
			}
			FDGGWSApiOrder port = getService().getFDGGWSApiOrderPort();
			setHandlers((BindingProvider) port);
			FDGGWSApiActionRequest request = new FDGGWSApiActionRequest();
			Action action = new Action();

			RecurringPayment recurringPayment = new RecurringPayment();

			recurringPayment.setFunction(CANCEL_FUNCTION);

			recurringPayment.setOrderId(getPreviousTransactionId());

			request.setAction(action);

			try {
				actionResponse = port.fdggwsApiAction(request);
			} catch (SOAPFaultException e) {
				String errMsg = "A soap fault occurred while trying to perform a refund transaction";
				log.log(Level.SEVERE, errMsg, e);
				throw new PaymentAuthorizationException(lastFaultMessage != null ? lastFaultMessage : errMsg);
			}
			if(!actionResponse.getTransactionResult().equals("APPROVED")) {
				throw new PaymentAuthorizationException(orderResponse.getErrorMessage(), getLastResponse());
			}
		}

		public void updateSubscription() throws PaymentAuthorizationException {
			createRecurringTransaction(MODIFY_FUNCTION, isEcheck());
		}

		public void createSubscription() throws PaymentAuthorizationException {
			createRecurringTransaction(null, isEcheck());
		}

		public void authCaptureAchTxn() throws PaymentAuthorizationException, PaymentIOException {
			try {
				initializeSecuritySettings();
			} catch (Exception e) {
				String errMsg = "An error occurred while initializing the client-side certificate settings";
				log.log(Level.SEVERE, errMsg, e);
			}
			FDGGWSApiOrder port = getService().getFDGGWSApiOrderPort();
			setHandlers((BindingProvider) port);
			FDGGWSApiOrderRequest orderRequest = new FDGGWSApiOrderRequest();

			Transaction transaction = new Transaction();

			createAuthAchTransaction(transaction, TransactionTypeEnum.AuthCaptureTxn);

			orderRequest.setTransaction(transaction);

			try {
				orderResponse = port.fdggwsApiOrder(orderRequest);
			} catch (SOAPFaultException e) {
				String errMsg = "A soap fault occurred while trying to perform a sale transaction";
				log.log(Level.SEVERE, errMsg, e);
				throw new PaymentAuthorizationException(lastFaultMessage != null ? lastFaultMessage : errMsg);
			}
			if(!orderResponse.getTransactionResult().equals("APPROVED")) {
				throw new PaymentAuthorizationException(orderResponse.getErrorMessage(), getLastResponse());
			}
		}

		public void createSubscriptionAch() throws PaymentAuthorizationException, PaymentIOException {
			createRecurringTransaction(null, true);
		}

		public void addPaymentToProfileAch() throws PaymentAuthorizationException, PaymentIOException {
			throw new UnsupportedOperationException("This operation is not supported");
		}

		public void updatePaymentProfileAch() throws PaymentAuthorizationException, PaymentIOException {
			throw new UnsupportedOperationException("This operation is not supported");
		}

		public void updateSubscriptionAch() throws PaymentAuthorizationException, PaymentIOException {
			createRecurringTransaction(MODIFY_FUNCTION, true);
		}

		public void cancelSubscriptionAch() throws PaymentAuthorizationException, PaymentIOException {
			createRecurringTransaction(CANCEL_FUNCTION, true);
		}

		public void authCaptureProfileAchTxn() throws PaymentAuthorizationException, PaymentIOException {
			throw new UnsupportedOperationException("This operation is not supported");
		}

		public String getAccountNumber() {
			return FirstDataGateway2.this.getAccountNumber();
		}

		public String getAccountName() {
			return FirstDataGateway2.this.getAccountName();
		}

		public String getAccountHolderName() {
			return FirstDataGateway2.this.getAccountHolderName();
		}

		public String getAbaCode() {
			return FirstDataGateway2.this.getAbaCode();
		}

		public Object getAccountType() {
			return FirstDataGateway2.this.getAccountType();
		}

		public String getBankName() {
			return FirstDataGateway2.this.getBankName();
		}

		public String getBankState() {
			return FirstDataGateway2.this.getBankState();
		}

		public String getCheckNumber() {
			return FirstDataGateway2.this.getCheckNumber();
		}

		public String getDriversLicenseNumber() {
			return FirstDataGateway2.this.getDriversLicenseNumber();
		}

		public String getDriversLicenseState() {
			return FirstDataGateway2.this.getDriversLicenseState();
		}

		public String getSsn() {
			return FirstDataGateway2.this.getSsn();
		}
	}

	@Override
	public PaymentInterface getPaymentInterface() {
		return new FirstDataPaymentInterface();
	}

	private void createRecurringTransaction(String function, boolean isECheck) throws PaymentAuthorizationException {
		try {
			initializeSecuritySettings();
		} catch (Exception e) {
			String errMsg = "An error occurred while initializing the client-side certificate settings";
			log.log(Level.SEVERE, errMsg, e);
		}
		FDGGWSApiOrder port = getService().getFDGGWSApiOrderPort();
		setHandlers((BindingProvider) port);
		FDGGWSApiActionRequest request = new FDGGWSApiActionRequest();
		Action action = new Action();

		RecurringPayment recurringPayment = new RecurringPayment();

		if(function != null) {
			recurringPayment.setFunction(function);
		}

		RecurringPaymentInformation recurringPaymentInformation = new RecurringPaymentInformation();

		recurringPaymentInformation.setRecurringStartDate(recurringFormat.format(getRecurringStartDate()));
		recurringPaymentInformation.setInstallmentCount(getRecurringNumberInstallments());
		recurringPaymentInformation.setInstallmentFrequency(getRecurringInstallmentFrequency());
		recurringPaymentInformation.setInstallmentPeriod(getRecurringInstallmentPeriod());
		recurringPaymentInformation.setMaximumFailures(getRecurringMaxFailures());

		recurringPayment.setRecurringPaymentInformation(recurringPaymentInformation);
		RecurringPayment.TransactionDataType transaction = new RecurringPayment.TransactionDataType();

		if(isECheck) {
			TeleCheckData checkData = new TeleCheckData();
			transaction.setTeleCheckData(checkData);
			checkData.setAccountNumber(getAccountNumber());
			checkData.setAccountType(getAccountType().getAccountTypeAbbreviation());
			checkData.setBankName(getBankName());
			checkData.setBankState(getBankState());
			checkData.setCheckNumber(getCheckNumber());
			checkData.setDrivingLicenseNumber(getDriversLicenseNumber());
			checkData.setDrivingLicenseState(getDriversLicenseState());
			checkData.setRoutingNumber(getAbaCode());
			checkData.setSocialSecurityNumber(getSsn());

			transaction.setTeleCheckData(checkData);
		} else {
			CreditCardData cardData = new CreditCardData();

			if(isCardPresent() && getSwipe() != null) {
				cardData.setTrackData(getSwipe());
			} else {
				cardData.setCardNumber(getCardNumber());
				cardData.setExpMonth(monthFormat.format(getExpirationDate()));
				cardData.setExpYear(yearFormat.format(getExpirationDate()));
			}

			if(getCardCodeIndicator() != null) {
				cardData.setCardCodeIndicator(getCardCodeIndicator().name());
			}
			cardData.setCardCodeValue(getCardVerificationValue());

			transaction.setCreditCardData(cardData);
		}

		Payment payment = new Payment();
		payment.setChargeTotal(getAmount());

		recurringPayment.setPayment(payment);

		TransactionDetails transactionDetails = new TransactionDetails();

		transactionDetails.setInvoiceNumber(getInvoiceNumber());
		transactionDetails.setOrderId(getOrderId());
		transactionDetails.setReferenceNumber(getReferenceId());
		transactionDetails.setTDate(gettDate());
		transactionDetails.setTransactionOrigin(isCardPresent() ? "RETAIL" : "ECI");
		transactionDetails.setPONumber(getPoNumber());

		recurringPayment.setTransactionDetails(transactionDetails);

		Billing billing = new Billing();
		billing.setCustomerID(getCustomerId());
		billing.setName(getFullName());
		billing.setCompany(getCompanyName());
		billing.setAddress1(getAddress());
		billing.setAddress2(getAddress2());
		billing.setCity(getCity());
		billing.setState(getState());
		billing.setZip(getZip());
		billing.setCountry(getCountry());
		billing.setPhone(getPhone());
		billing.setFax(getFax());
		billing.setEmail(getEmail());

		recurringPayment.setBilling(billing);

		Shipping shipping = new Shipping();

		shipping.setAddress1(getShipAddress());
		shipping.setAddress2(getShipAddress2());
		shipping.setCity(getShipCity());
		shipping.setState(getShipState());
		shipping.setZip(getShipZip());
		shipping.setCountry(getShipCountry());

		recurringPayment.setShipping(shipping);

		Notes notes = new Notes();
		notes.setComments(getChargeDescription());
		recurringPayment.setNotes(notes);

		recurringPayment.setTransactionDataType(transaction);

		action.setRecurringPayment(recurringPayment);

		request.setAction(action);

		try {
			actionResponse = port.fdggwsApiAction(request);
		} catch (SOAPFaultException e) {
			String errMsg = "A soap fault occurred while trying to perform a refund transaction";
			log.log(Level.SEVERE, errMsg, e);
			throw new PaymentAuthorizationException(lastFaultMessage != null ? lastFaultMessage : errMsg);
		}
		if(!actionResponse.getTransactionResult().equals("APPROVED")) {
			throw new PaymentAuthorizationException(orderResponse.getErrorMessage(), getLastResponse());
		}
	}


	private void createAuthTransaction(Transaction transaction, TransactionTypeEnum transactionType) {
		CreditCardTxType txType = new CreditCardTxType();

		txType.setType(transactionType.getTransactionType());

		transaction.setCreditCardTxType(txType);

		CreditCardData cardData = new CreditCardData();

		if(isCardPresent() && getSwipe() != null) {
			cardData.setTrackData(getSwipe());
		} else {
			cardData.setCardNumber(getCardNumber());
			cardData.setExpMonth(monthFormat.format(getExpirationDate()));
			cardData.setExpYear(yearFormat.format(getExpirationDate()));
		}

		if(getCardCodeIndicator() != null) {
			cardData.setCardCodeIndicator(getCardCodeIndicator().name());
		}
		cardData.setCardCodeValue(getCardVerificationValue());

		transaction.setCreditCardData(cardData);

		Payment payment = new Payment();
		payment.setChargeTotal(getAmount());

		transaction.setPayment(payment);

		TransactionDetails transactionDetails = new TransactionDetails();

		transactionDetails.setInvoiceNumber(getInvoiceNumber());
		transactionDetails.setOrderId(getOrderId());
		transactionDetails.setReferenceNumber(getReferenceId());
		transactionDetails.setTDate(gettDate());
		transactionDetails.setTransactionOrigin(isCardPresent() ? "RETAIL" : "ECI");
		transactionDetails.setPONumber(getPoNumber());

		transaction.setTransactionDetails(transactionDetails);

		Billing billing = new Billing();
		billing.setCustomerID(getCustomerId());
		billing.setName(getFullName());
		billing.setCompany(getCompanyName());
		billing.setAddress1(getAddress());
		billing.setAddress2(getAddress2());
		billing.setCity(getCity());
		billing.setState(getState());
		billing.setZip(getZip());
		billing.setCountry(getCountry());
		billing.setPhone(getPhone());
		billing.setFax(getFax());
		billing.setEmail(getEmail());

		transaction.setBilling(billing);

		Shipping shipping = new Shipping();

		shipping.setAddress1(getShipAddress());
		shipping.setAddress2(getShipAddress2());
		shipping.setCity(getShipCity());
		shipping.setState(getShipState());
		shipping.setZip(getShipZip());
		shipping.setCountry(getShipCountry());

		transaction.setShipping(shipping);

		Notes notes = new Notes();
		notes.setComments(getChargeDescription());
		transaction.setNotes(notes);
	}

	private void createAuthAchTransaction(Transaction transaction, TransactionTypeEnum transactionType) {
		TeleCheckTxType txType = new TeleCheckTxType();
		txType.setType(transactionType.getTransactionType());

		TeleCheckData checkData = new TeleCheckData();
		transaction.setTeleCheckData(checkData);
		checkData.setAccountNumber(getAccountNumber());
		checkData.setAccountType(getAccountType().name());
		checkData.setBankName(getBankName());
		checkData.setBankState(getBankState());
		checkData.setCheckNumber(getCheckNumber());
		checkData.setDrivingLicenseNumber(getDriversLicenseNumber());
		checkData.setDrivingLicenseState(getDriversLicenseState());
		checkData.setRoutingNumber(getAbaCode());
		checkData.setSocialSecurityNumber(getSsn());

		transaction.setTeleCheckData(checkData);

		Payment payment = new Payment();
		payment.setChargeTotal(getAmount());

		transaction.setPayment(payment);

		TransactionDetails transactionDetails = new TransactionDetails();

		transactionDetails.setInvoiceNumber(getInvoiceNumber());
		transactionDetails.setOrderId(getOrderId());
		transactionDetails.setReferenceNumber(getReferenceId());
		transactionDetails.setTDate(gettDate());
		transactionDetails.setTransactionOrigin(isCardPresent() ? "RETAIL" : "ECI");
		transactionDetails.setPONumber(getPoNumber());

		transaction.setTransactionDetails(transactionDetails);

		Billing billing = new Billing();
		billing.setCustomerID(getCustomerId());
		billing.setName(getFullName());
		billing.setCompany(getCompanyName());
		billing.setAddress1(getAddress());
		billing.setAddress2(getAddress2());
		billing.setCity(getCity());
		billing.setState(getState());
		billing.setZip(getZip());
		billing.setCountry(getCountry());
		billing.setPhone(getPhone());
		billing.setFax(getFax());
		billing.setEmail(getEmail());

		transaction.setBilling(billing);

		Shipping shipping = new Shipping();

		shipping.setAddress1(getShipAddress());
		shipping.setAddress2(getShipAddress2());
		shipping.setCity(getShipCity());
		shipping.setState(getShipState());
		shipping.setZip(getShipZip());
		shipping.setCountry(getShipCountry());

		transaction.setShipping(shipping);

		Notes notes = new Notes();
		notes.setComments(getChargeDescription());
		transaction.setNotes(notes);
	}

	private void setHandlers(BindingProvider bindingProvider) {
		bindingProvider.getRequestContext().put(BindingProvider.USERNAME_PROPERTY, getApiMerchantAccountName());
		bindingProvider.getRequestContext().put(BindingProvider.PASSWORD_PROPERTY, getApiPasswordOrKey());
		if(getUrl() != null) {
			bindingProvider.getRequestContext().put(BindingProvider.ENDPOINT_ADDRESS_PROPERTY, getUrl());
		} else if(!isTestMode()) {
			bindingProvider.getRequestContext().put(BindingProvider.ENDPOINT_ADDRESS_PROPERTY, PRODUCTION_URL);
		}
		Binding binding = bindingProvider.getBinding();
		List<Handler> handlers = binding.getHandlerChain();
		LoggingHandler handler = new LoggingHandler() {
			@Override
			public void setLastInboundMsg(String lastInboundMsg) {
				super.setLastInboundMsg(lastInboundMsg);
				setLastResponse(lastInboundMsg);
			}
		};
		handlers.add(handler);
		handlers.add(new SoapFaultEditingHandler() {
			@Override
			public void setLastFaultMessage(String lastFaultMessage) {
				super.setLastFaultMessage(lastFaultMessage);
				FirstDataGateway2.this.lastFaultMessage = lastFaultMessage;
			}
		});
		binding.setHandlerChain(handlers);
	}

	private String parseSoapFault(String soapFault) throws ParserConfigurationException, IOException, SAXException {
		StringBuilder result = new StringBuilder("");
		DocumentBuilder builder = DocumentBuilderFactory.newInstance().newDocumentBuilder();
		Document document = builder.parse(new InputSource(new StringReader(soapFault.trim())));
		NodeList faultString = document.getElementsByTagName("faultString");
		result.append(faultString != null ? faultString : "");
		NodeList detail = document.getElementsByTagName("detail");
		result.append(detail != null ? detail : "");
		return result.toString();
	}

	@Override
	public PaymentResponse getResponseInterface() {
		return new CreditCardResponse() {
			public String getAvsResponse() {
				return orderResponse.getAVSResponse();
			}

			public String getCvvResponse() {
				return "";
			}

			public String getTransactionId() {
				if(transactionType.equals(TransactionTypeEnum.AuthCaptureTxn) || transactionType.equals(TransactionTypeEnum.CaptureTxn)) {
					return orderResponse.getOrderId() + "|" + orderResponse.getTDate();
				} else {
					return orderResponse.getOrderId();
				}
			}

			public String getTransactionResult() {
				return orderResponse.getTransactionResult();
			}

			public String getRawResponse() {
				return getLastResponse() == null ? "" : getLastResponse();
			}

			public String getAuthorizationCode() {
				return orderResponse.getApprovalCode();
			}
		};
	}

	public String getAbaCode() {
		return abaCode;
	}

	public String getAccountHolderName() {
		return accountHolderName;
	}

	public String getAccountName() {
		return accountName;
	}

	public String getAccountNumber() {
		return accountNumber;
	}

	public AccountType getAccountType() {
		return accountType;
	}

	public String getBankName() {
		return bankName;
	}

	public String getBankState() {
		return bankState;
	}

	public String getCheckNumber() {
		return checkNumber;
	}

	public String getDriversLicenseNumber() {
		return driversLicenseNumber;
	}

	public String getDriversLicenseState() {
		return driversLicenseState;
	}

	public String getSsn() {
		return ssn;
	}

	public void setAbaCode(String abaCode) {
		this.abaCode = abaCode;
	}

	public void setAccountHolderName(String accountHolderName) {
		this.accountHolderName = accountHolderName;
	}

	public void setAccountName(String accountName) {
		this.accountName = accountName;
	}

	public void setAccountNumber(String accountNumber) {
		this.accountNumber = accountNumber;
	}

	public void setAccountType(AccountType accountType) {
		this.accountType = accountType;
	}

	public void setActionResponse(FDGGWSApiActionResponse actionResponse) {
		this.actionResponse = actionResponse;
	}

	public void setBankName(String bankName) {
		this.bankName = bankName;
	}

	public void setBankState(String bankState) {
		this.bankState = bankState;
	}

	public void setCertificatePassword(String certificatePassword) {
		this.certificatePassword = certificatePassword;
	}

	public void setCertificatePath(String certificatePath) {
		this.certificatePath = certificatePath;
	}

	public void setCheckNumber(String checkNumber) {
		this.checkNumber = checkNumber;
	}

	public void setDriversLicenseNumber(String driversLicenseNumber) {
		this.driversLicenseNumber = driversLicenseNumber;
	}

	public void setDriversLicenseState(String driversLicenseState) {
		this.driversLicenseState = driversLicenseState;
	}

	public void setLastFaultMessage(String lastFaultMessage) {
		this.lastFaultMessage = lastFaultMessage;
	}

	public void setOrderResponse(FDGGWSApiOrderResponse orderResponse) {
		this.orderResponse = orderResponse;
	}

	public static void setService(FDGGWSApiOrderService service) {
		FirstDataGateway2.service = service;
	}

	public void setSsn(String ssn) {
		this.ssn = ssn;
	}
}
